;DCF77 receiver with plausability check, automatic summer/winter time switching
;and UTC time calculation for ATTiny2313 @ 4.194304 MHz
;
;*** Include this file after interrupt vectors (if any) in main program ***
;call "dcf_init" before entering main loop, don't overwrite TIMSK afterwards.
;
;REGISTERS:
;  shared registers (R16-R27): temp, temp2
;  static registers (R16-R27): dcfbits, dcfflags
;  clock registers (R0-R15): seconds (cleared on successful reception)
;
;CONSTANTS:
;  DCF_INVERTED - 0 for non-inverted (high-pulses), 1 for inverted (low-pulses)
;  DCF_USE_INT0 - 0 to use ICP1 (PD6) as DCF77 input, 1 to use INT0 (PD2)
;
;TIMERS:
;  TIMER1 cannot be used for anything else except DCF77 reception.
;  Connect DCF77 signal to ICP1 pin (PD6).
;
;TIME AND DATE IN RAM:
;  RAM_Time: contains the local time and date in packed BCD format
;  RAM_UTC: contains the UTC time and date in packed BCD format
;    OFFSET   CONTENTS
;        0    Minutes (0x00 to 0x59)
;        1    Hours (0x00 to 0x23)
;        2    Day of Week (1 = Monday to 7 = Sunday)
;        3    Day of Month (0x01 to 0x31)
;        4    Month (0x01 to 0x12)
;        5    2-Digit Year (0x00 to 0x99)
;        6    Flags (see below)
;  when incrementing time, call "inctime", then immediately call "calculate_utc"

;VERSION HISTORY:
;  2010-02-08  Arne Rossius
;    (first version)
;  2010-02-26  Arne Rossius
;    * added RAM offset constants for Minutes, Hours etc.

;TODO: interpret double pulse as one long pulse

.equ DCF77_STANDALONE = 0 ;set to 1 to assemble this stand-alone (debugging)

;===============================================================================

;Flags for 'dcfflags' register:
.equ	dcfErrBitLength = 0x1 ;invalid bit length
.equ	dcfErrBitCount = 0x2 ;wrong number of bits
.equ	dcfErrParity = 0x3 ;parity error
.equ	dcfErrStartBit = 0x4 ;start bit wasn't 1
.equ	dcfErrPlausibility = 0x5 ;plausibility check failed
.equ	dcfErrDST = 0x6 ;UTC+1 and UTC+2 (DST) both set or both unset
.equ	dcfErrMask = 0x07 ;(used to reset errors)
.equ	dcfTimeOK = 3 ;time is OK
.equ	dcfStop = 4 ;stop DCF77 reception until a new telegram starts
.equ	dcfTime = 5 ;Last synchronisation attempt was successful
.equ	dcfFirstTelegram = 6 ;first telegram received (compare to next)

.if (DCF77_STANDALONE != 0)
	.warning "DCF77 RECEIVER: STAND-ALONE OPTION SET!"
	.include "../include/tn2313def.inc"
	.equ DCF_INVERTED = 1
	.equ DCF_USE_INT0 = 0
	.def seconds = R0
	.def temp = R16
	.def temp2 = R17
	.def dcfbits = R18
	.def dcfflags = R19
.endif

;===============================================================================

.dseg
	RAM_Time: .byte 7 ;(minutes, hours, weekday, day of month, month, year, flags)
	RAM_UTC: .byte 7
	RAM_DCF1: .byte 7
	RAM_DCF2: .byte 7
		;FLAGS:
		.equ	timeDST = 0 ;daylight saving time
	
.equ	tMinute = 0
.equ	tHour = 1
.equ	tWeekday = 2
.equ	tDay = 3
.equ	tMonth = 4
.equ	tYear = 5
.equ	tFlags = 6

;===============================================================================

.cseg
.if (DCF_USE_INT0 == 0)
.org ICP1addr
.else
.org INT0addr
.endif
	rjmp	dcf_capture
.org OVF1addr
	rjmp	dcf_overflow
.org INT_VECTORS_SIZE

;===============================================================================

dcf_init:
	;initialize dcf77 registers, RAM and Timer1
	clr	dcfbits
	ldi	dcfflags, 1<<dcfStop
	
	ldi	temp, 0x00
	ldi	YL, RAM_Time
dcf_init_ram:
	st	Y+, temp
	cpi	YL, RAM_Time+4*7
	brlo	dcf_init_ram
	
.if (DCF_USE_INT0 == 0)
	.if (DCF_INVERTED == 0)
	ldi	temp, 1<<ICNC1 | 1<<ICES1 | 0x05 ;Clk / 1024 (ICP: pos. edge)
	out	TCCR1B, temp
	.else
	ldi	temp, 1<<ICNC1 | 0<<ICES1 | 0x05 ;Clk / 1024 (ICP: neg. edge)
	out	TCCR1B, temp
	.endif
	in	temp, TIMSK
	ori	temp, 1<<ICIE1 | 1<<TOIE1
	out	TIMSK, temp
.else
	.if (DCF_INVERTED == 0)
	ldi	temp, 1<<ISC01 | 1<<ISC00 ;INT0: positive edge
	out	MCUCR, temp
	.else
	ldi	temp, 1<<ISC01 ;INT0: negative edge
	out	MCUCR, temp
	.endif
	ldi	temp, 1<<INT0
	out	GIMSK, temp
	ldi	temp, 0x05 ;Clk/1024
	out	TCCR1B, temp
	in	temp, TIMSK
	ori	temp, 1<<TOIE1
	out	TIMSK, temp
.endif
	
	ret

;===============================================================================

dcf_overflow:
	;DCF77 timer overflow: no change for 16 seconds!
	;reset timer to 8192 (= 2s), no SREG backup needed
	push	temp
	ldi	temp, HIGH(8192)
	out	TCNT1H, temp
	ldi	temp, LOW(8192)
	out	TCNT1L, temp
	pop	temp
	reti

;===============================================================================

dcf_capture:
	;DCF77 input capture interrupt: DCF77 signal edge detected
	push	temp
	push	temp2
	push	YL
	in	temp, SREG
	push	temp
	
;---------------------------------------
; DCF77 reception
;---------------------------------------
	
.if (DCF_USE_INT0 == 0)
	;reset timer
	ldi	temp, 0
	out	TCNT1H, temp
	out	TCNT1L, temp
	;get time since last DCF77 signal edge
	in	temp, ICR1L
	in	temp2, ICR1H
	;positive or negative edge?
	.if (DCF_INVERTED == 0)
	in	YL, TCCR1B
	sbrc	YL, ICES1
	rjmp	dcf_posedge
	ori	YL, 1<<ICES1 ;next edge will be positive
	out	TCCR1B, YL
	.else
	in	YL, TCCR1B
	sbrs	YL, ICES1
	rjmp	dcf_posedge
	andi	YL, ~(1<<ICES1) ;next edge will be negative
	out	TCCR1B, YL
	.endif
.else
	;get time since last DCF77 signal edge
	in	temp, TCNT1L
	in	temp2, TCNT1H
	;reset timer
	ldi	YL, 0
	out	TCNT1H, YL
	out	TCNT1L, YL
	;positive or negative edge?
	.if (DCF_INVERTED == 0)
	in	YL, MCUCR
	sbrc	YL, ISC00
	rjmp	dcf_posedge
	ori	YL, 1<<ISC00 ;next edge will be positive
	out	MCUCR, YL
	.else
	in	YL, MCUCR
	sbrs	YL, ISC00
	rjmp	dcf_posedge
	andi	YL, ~(1<<ISC00) ;next edge will be negative
	out	MCUCR, YL
	.endif
.endif
	
	;negative edge: one bit received (inverted signal: positive edge)
	sbrc	dcfflags, dcfStop ;ignore if reception stopped
	rjmp	dcf_end
	ldi	YL, HIGH(205) ;50 ms
	cpi	temp, LOW(205)
	cpc	temp2, YL
	brlo	dcf_err_bitlen
	ldi	YL, HIGH(614) ;150 ms
	cpi	temp, LOW(614)
	cpc	temp2, YL
	brlo	dcf_bit_zero ;50ms <= bit length < 150 ms  -->  bit = 0
	ldi	YL, HIGH(1024) ;250 ms
	cpi	temp, LOW(1024)
	cpc	temp2, YL
	brlo	dcf_bit_one ;150ms <= bit length < 250 ms  -->  bit = 1
	
dcf_err_bitlen:
	;invalid bit length, stop reception for this telegram
	andi	dcfflags, ~(dcfErrMask | 1<<dcfTime | 1<<dcfFirstTelegram)
	ori	dcfflags, dcfErrBitLength | 1<<dcfStop
	rjmp	dcf_end
	
dcf_bit_zero:
	clc ;clear bit
	rjmp	dcf_bit_save
dcf_bit_one:
	sec ;set bit
dcf_bit_save:
	ldi	YL, RAM_DCF2+6
	ld	temp, -Y
	ror	temp
	st	Y, temp
	ld	temp, -Y
	ror	temp
	st	Y, temp
	ld	temp, -Y
	ror	temp
	st	Y, temp
	ld	temp, -Y
	ror	temp
	st	Y, temp
	ld	temp, -Y
	ror	temp
	st	Y, temp
	ld	temp, -Y
	ror	temp
	st	Y, temp
	inc	dcfbits
	cpi	dcfbits, 60 ;stop reception if telegram contains more than 59 bits
	brlo	dcf_end
	andi	dcfflags, ~(dcfErrMask | 1<<dcfTime | 1<<dcfFirstTelegram)
	ori	dcfflags, dcfErrBitCount | 1<<dcfStop
	rjmp	dcf_end

dcf_posedge:
	;positive edge: beginning of a new bit (inverted signal: negative edge)
.if (DCF_USE_INT0 == 0)
	.if (DCF_INVERTED == 0)
	andi	YL, ~(1<<ICES1) ;next edge will be negative
	out	TCCR1B, YL
	.else
	ori	YL, 1<<ICES1 ;next edge will be positive
	out	TCCR1B, YL
	.endif
.else
	.if (DCF_INVERTED == 0)
	andi	YL, ~(1<<ISC00) ;next edge will be negative
	out	MCUCR, YL
	.else
	ori	YL, 1<<ISC00 ;next edge will be positive
	out	MCUCR, YL
	.endif
.endif
	ldi	YL, HIGH(6144) ;1.5s
	cpi	temp, LOW(6144)
	cpc	temp2, YL
	brsh	dcf_complete ;pause >= 1.5s: telegram complete
	
dcf_end:
	pop	temp
	out	SREG, temp
	pop	YL
	pop	temp2
	pop	temp
	reti

;---------------------------------------
; DCF77 decoding
;---------------------------------------

dcf_complete:
	;end of telegram / new telegram
	mov	temp, dcfflags
	mov	temp2, dcfbits
	andi	dcfflags, ~(dcfErrMask | 1<<dcfStop)
	ldi	dcfbits, 0
	sbrc	temp, dcfStop ;don't decode if DCF77 reception was stopped
	rjmp	dcf_end
	cpi	temp2, 59 ;check if correct number of bits were received
	breq	dcf_decode
	andi	dcfflags, ~(1<<dcfTime)
	ori	dcfflags, dcfErrBitCount
	rjmp	dcf_end
	
dcf_decode:
	;check startbit
	lds	temp, RAM_DCF2+1
	sbrs	temp, 1
	rjmp	dcf_err_startbit
	
	;check minute parity
	ldi	temp2, 0
	andi	temp, 0xFC
	rcall	parity
	lds	temp, RAM_DCF2+2
	mov	YL, temp
	andi	temp, 0x03
	rcall	parity
	sbrc	temp2, 0
	rjmp	dcf_err_parity

	;check hour parity
	ldi	temp2, 0
	mov	temp, YL
	andi	temp, 0xFC
	rcall	parity
	lds	temp, RAM_DCF2+3
	sbrc	temp, 0
	inc	temp2
	sbrc	temp2, 0
	rjmp	dcf_err_parity
	
	;check date parity
	ldi	temp2, 0
	andi	temp, 0xFE
	rcall	parity
	lds	temp, RAM_DCF2+4
	rcall	parity
	lds	temp, RAM_DCF2+5
	rcall	parity
	sbrc	temp2, 0
	rjmp	dcf_err_parity
	
	;save DST
	lds	temp, RAM_DCF2
	andi	temp, 0xC0
	breq	PC+3 ;zero: error
	cpi	temp, 0xC0
	brne	PC+2 ;both bits set: error
	rjmp	dcf_err_dst
	ldi	temp2, 0
	sbrc	temp, 6
	ldi	temp2, 1<<timeDST
	sts	RAM_DCF2+6, temp2
	
	;save minutes
	lds	temp, RAM_DCF2+1
	lsr	YL
	ror	temp
	lsr	temp
	sts	RAM_DCF2+tMinute, temp
	
	;save hours
	lsr	YL
	sts	RAM_DCF2+tHour, YL
	
	;save day
	lds	temp, RAM_DCF2+3
	lds	temp2, RAM_DCF2+4
	lds	YL, RAM_DCF2+5
	lsl	temp ;align day
	rol	temp2 ;align month[5] weekday[3]
	rol	YL ;align year
	lsr	temp
	lsr	temp
	sts	RAM_DCF2+tDay, temp
	
	;save weekday
	mov	temp, temp2
	andi	temp, 0x07
	sts	RAM_DCF2+tWeekday, temp
	
	;save month
	lsr	temp2
	lsr	temp2
	lsr	temp2
	sts	RAM_DCF2+tMonth, temp2
	
	;save year
	sts	RAM_DCF2+tYear, YL
	
	;check time/date against previous reception
	sbrs	dcfflags, dcfFirstTelegram ;don't compare if this is the first telegram
	rjmp	dcf_copy_invalid
	ldi	YL, RAM_DCF1
	rcall	inctime
dcf_check:
	ldd	temp2, Y+(RAM_DCF2 - RAM_DCF1)
	ld	temp, Y+
	cp	temp, temp2
	brne	dcf_check_failed
	cpi	YL, RAM_DCF1+7
	brlo	dcf_check
	;plausibility check succeeded, copy new time to old and to valid time
	ldi	YL, RAM_Time
dcf_copy_valid:
	ldd	temp, Y+(RAM_DCF2 - RAM_Time)
	std	Y+(RAM_DCF1 - RAM_Time), temp
	st	Y+, temp
	cpi	YL, RAM_Time+7
	brlo	dcf_copy_valid
	ori	dcfflags, 1<<dcfTimeOK | 1<<dcfTime
	clr	seconds
	rcall	calculate_utc
	rjmp	dcf_copy_end
dcf_check_failed:
	;plausibility check failed: only copy new time to old time
	andi	dcfflags, ~(1<<dcfTime)
	ori	dcfflags, dcfErrPlausibility
dcf_copy_invalid:
	ori	dcfflags,1<<dcfFirstTelegram
	ldi	YL, RAM_DCF1
dcf_copy_invalid_loop:
	ldd	temp, Y+(RAM_DCF2 - RAM_DCF1)
	st	Y+, temp
	cpi	YL, RAM_DCF1+7
	brne	dcf_copy_invalid_loop
dcf_copy_end:
	rjmp	dcf_end
	
dcf_err_parity:
	;set parity error
	andi	dcfflags, ~(1<<dcfTime | 1<<dcfFirstTelegram)
	ori	dcfflags, dcfErrParity
	rjmp	dcf_end

dcf_err_startbit:
	;set start-bit error
	andi	dcfflags, ~(1<<dcfTime | 1<<dcfFirstTelegram)
	ori	dcfflags, dcfErrStartBit
	rjmp	dcf_end

dcf_err_dst:
	;summer and winter time are both set or both unset
	andi	dcfflags, ~(1<<dcfTime | 1<<dcfFirstTelegram)
	ori	dcfflags, dcfErrDST
	rjmp	dcf_end

;===============================================================================

parity:
	;calculate parity (add to temp2)
	sbrc	temp, 0
	inc	temp2
	lsr	temp
	brne	parity
	ret
	
;--------------------

inctime:
	;increment time at RAM[Y] by 1 minute (without modifying Y)
	ldd	temp, Y+tMinute ;minutes
	rcall	bcdinc
	st	Y, temp
	cpi	temp, 0x60
	brlo	inctime_end
	ldi	temp, 0x00
	std	Y+tMinute, temp
	ldd	temp, Y+tHour ;hours
	rcall	bcdinc
	std	Y+tHour, temp
	cpi	temp, 0x24
	brsh	inctime_day
	
	;DST auto-switching (enable)
	ldd	temp2, Y+tFlags
	sbrc	temp2, timeDST
	rjmp	inctime_dst
	cpi	temp, 0x02 ;last sunday in march, 02:00 => set clock to 03:00
	brne	inctime_end
	ldd	temp, Y+tWeekday ;day of week = 7 ?
	cpi	temp, 7
	brne	inctime_end
	ldd	temp, Y+tDay ;day of month = 25~31 ?
	cpi	temp, 0x25
	brlo	inctime_end
	ldd	temp, Y+tMonth ;month = 3 ?
	cpi	temp, 0x03
	brne	inctime_end
	ldi	temp, 0x03
	std	Y+tHour, temp
	ori	temp2, 1<<timeDST
	std	Y+tFlags, temp2
	rjmp	inctime_end
	
inctime_day:
	ldi	temp, 0x00
	std	Y+tHour, temp
	ldd	temp, Y+tWeekday ;weekday
	inc	temp
	sbrc	temp, 3
	ldi	temp, 1
	std	Y+tWeekday, temp
	rcall	calculate_month_length
	ldd	temp, Y+tDay ;day
	rcall	bcdinc
	std	Y+tDay, temp
	cp	temp2, temp
	brsh	inctime_end
	ldi	temp, 0x01
	std	Y+tDay, temp
	ldd	temp, Y+tMonth ;month
	rcall	bcdinc
	std	Y+tMonth, temp
	cpi	temp, 0x13
	brlo	inctime_end
	ldi	temp, 0x01
	std	Y+tMonth, temp
	ldd	temp, Y+tYear ;year
	rcall	bcdinc
	std	Y+tYear, temp
	cpi	temp, 0x9A
	brlo	inctime_end
	ldi	temp, 0x00
	std	Y+tYear, temp
inctime_end:
	ret

inctime_dst:
	;DST auto-switching (disable)
	cpi	temp, 0x03 ;last sunday in october, 03:00 => set clock to 02:00
	brne	inctime_end
	ldd	temp, Y+tWeekday ;day of week = 7 ?
	cpi	temp, 7
	brne	inctime_end
	ldd	temp, Y+tDay ;day of month = 25~31 ?
	cpi	temp, 0x25
	brlo	inctime_end
	ldd	temp, Y+tMonth ;month = 10 ?
	cpi	temp, 0x10
	brne	inctime_end
	ldi	temp, 0x02
	std	Y+tHour, temp
	andi	temp2, ~(1<<timeDST)
	std	Y+tFlags, temp2
	rjmp	inctime_end

;--------------------

calculate_month_length:
	;calculate month length (result goes into temp2)
	;month is taken from RAM[Y+4], year from RAM[Y+5]
	ldd	temp, Y+4
	ldi	temp2, 0x30
	cpi	temp, 0x08
	brlo	PC+2
	inc	temp
	sbrc	temp, 0
	ldi	temp2, 0x31
	cpi	temp, 0x02
	brne	calculate_month_length_end
	ldi	temp2, 0x28
	ldd	temp, Y+5
	andi	temp, 0x03
	brne	calculate_month_length_end
	ldi	temp2, 0x29
calculate_month_length_end:
	ret

;--------------------

calculate_utc:
	;calculate UTC from local time
	;copy local time to UTC time
 	ldi	YL, RAM_Time+7
utc_copy:
	ld	temp, -Y
	std	Y+(RAM_UTC - RAM_Time), temp
	cpi	YL, RAM_Time
	brne	utc_copy
	ldi	YL, RAM_UTC
	;decrement hours by 1 (winter time) or 2 (summer time)
	ldd	temp, Y+tHour ;hours
	rcall	bcddec
	ldd	temp2, Y+tFlags
	sbrc	temp2, timeDST
	rcall	bcddec
	std	Y+tHour, temp
	cpi	temp, 0x24
	brlo	utc_end
	subi	temp, (0xF9-0x23)
	std	Y+tHour, temp
	;decrement day of week and day of month
	ldd	temp, Y+tWeekday ;day of week
	dec	temp
	cpi	temp, 0
	brne	PC+2
	ldi	temp, 7
	std	Y+tWeekday, temp
	ldd	temp, Y+tDay ;day of month
	rcall	bcddec
	std	Y+tDay, temp
	cpi	temp, 0x00
	brne	utc_end
	ldd	temp, Y+tMonth ;month
	rcall	bcddec
	std	Y+tMonth, temp
	cpi	temp, 0x00
	brne	utc_month_length
	ldi	temp, 0x12
	std	Y+tMonth, temp
	ldd	temp, Y+tYear ;year
	rcall	bcddec
	cpi	temp, 0xF0
	brlo	PC+2
	ldi	temp, 0x99
	std	Y+tYear, temp
utc_month_length:
	rcall	calculate_month_length
	std	Y+tDay, temp2
utc_end:
	ret

;--------------------

bcdinc:
	;increment packed BCD number
	subi	temp, -7
	brhc	PC+2
	subi	temp, 6
	ret
	
;--------------------

bcddec:
	;decrement packed BCD number
	subi	temp, 1
	brhc	PC+2
	subi	temp, 6
	ret
	
;--------------------

