;Nixie Clock (with new DCF77 receiver)
;Controller: ATTiny2313 @ 4.194304 MHz
;
;Assemble with avra (http://avra.sourceforge.net/) only, this file contains
;directives specific for this assembler!
;
;(C) Arne Rossius, http://www.elektronik.de.vu
;
;PORTS:
;  PB6 = Shift Register Data (MSB first, seconds first, packed BCD)
;  PB7 = Shift Register Clock
;  PD2 = DCF77 receiver
;  PD3 = "Show Date" key/switch
;
;VERSION HISTORY:
;  0.1  2005-09-23  Arne Rossius
;  0.2  2007-05-20  Arne Rossius
;    * auto-increment date
;  0.3  2008-01-12  Arne Rossius
;    * date format selectable (dd-mm-yy or yy-mm-dd)
;  0.4  2008-03-15  Arne Rossius
;    * auto-increment weekday (for future use)
;    * disable DCF77 reception after 4 minutes without success
;  0.5  2008-04-22  Arne Rossius
;    * BUGFIX: auto-increment date: load ZH:ZL instead of YH:ZL (narf!)
;  0.6  2010-02-26  Arne Rossius
;    * new and improved DCF77 receiver (with plausibility check)
;  0.7  2010-04-18  Arne Rossius
;    * BUGFIX: display didn't return to time after showing date

;microcontroller definition file
.include "../include/2313def.inc"

;Nixie clock options
.equ	DATE_TIMEOUT = 5 ;seconds
.equ	DD_MM_YY = 0 ;German date format instead of international

;time/date registers
.def	oldout1 = R13
.def	oldout2 = R14
.def	oldout3 = R15

;general registers
.def	temp = R16
.def	temp2 = R17

;DCF77-specific registers
.def	dcfbits = R18
.def	dcfflags = R19

;nixie clock registers
.def	newout1 = R20
.def	newout2 = R21
.def	newout3 = R22
.def	datecnt = R23
.def	subsec = R24
.def	seconds = R25

;*******************************************************************************
;*
;*   MAIN PROGRAM
;*
;*******************************************************************************

;---------------------------------------
; Interrupt vectors
;---------------------------------------

.cseg

;Reset
.org $0000
	rjmp	start
	
.org OVF0addr
	rjmp	ovf0int
	
;---------------------------------------
; DCF77 receiver
;---------------------------------------

.equ DCF_USE_INT0 = 1
.equ DCF_INVERTED = 1
.include "dcf77.asm"

;---------------------------------------
; Timer 0 Overflow: 16 Hz interrupt
;---------------------------------------

ovf0int:
	push	temp
	push	temp2
	push	YL
	in	temp, SREG
	push	temp
	
	;increment time
	inc	subsec ;sub-seconds
	cpi	subsec, 16
	brlo	ovf0int_end
	clr	subsec
	cpi	datecnt, 0 ;date display timeout
	breq	PC+2
	dec	datecnt
	mov	temp, seconds ;seconds
	rcall	bcdinc
	mov	seconds, temp
	cpi	temp, 0x60
	brlo	ovf0int_end
	clr	seconds
	ldi	YL, RAM_Time ;everything else (minutes ... year)
	rcall	inctime
	rcall	calculate_utc
	
ovf0int_end:
	pop	temp
	out	SREG, temp
	pop	YL
	pop	temp2
	pop	temp
	reti

;---------------------------------------
; Initialization
;---------------------------------------
	
start:
	;stack pointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;disable analog comparator
	ldi	temp, 0x80
	out	ACSR, temp
	
	;I/O ports
	ldi	temp, 0xC0
	out	DDRB, temp
	ldi	temp, 0x3F
	out	PORTB, temp
	ldi	temp, 0x00
	out	DDRD, temp
	ldi	temp, 0x7F
	out	PORTD, temp
	
	;Timer 0
	ldi	temp, 0x05 ;Clk/1024 => 16 Hz overflow
	out	TCCR0, temp
	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	
	;registers
	ldi	datecnt, 0
	clr	oldout1
	clr	oldout2
	clr	oldout3
	ldi	newout1, 0x88
	ldi	newout2, 0x88
	ldi	newout3, 0x88
	ldi	subsec, 0
	ldi	seconds, 0
	
	;DCF77
	rcall	dcf_init
	
	;enable interrupts
	sei

;---------------------------------------
; Main loop
;   find out what must be displayed and
;   shift it out if the current display
;   is different
;---------------------------------------

loop:
	;check key
	sbic	PIND, 3
	rjmp	key_off
	;key is pressed
	cpi	datecnt, 0
	brne	check_display
	ldi	datecnt, DATE_TIMEOUT*2
	rjmp	check_display
key_off:
	;key is not pressed
	cpi	datecnt, DATE_TIMEOUT+1
	brsh	check_display
	clr	datecnt
	
check_display:
	sbrs	dcfflags, dcfTimeOK
	rjmp	info
	;display date or time?
	cpi	datecnt, DATE_TIMEOUT+1
	brsh	date
	rjmp	time
	
info:
	;display DCF77 reception info (88:x8:yy, x=error, yy=received bits)
	ldi	newout1, 0x88
	ldi	newout2, 0x08
	mov	temp, dcfflags
	andi	temp, dcfErrMask
	swap	temp
	or	newout2, temp
	mov	temp, dcfbits
	rcall	bin2bcd
	mov	newout3, temp
	
display:
	;check if the display has changed:
	cp	oldout1, newout1
	cpc	oldout2, newout2
	cpc	oldout3, newout3
	breq	loop
	;display the new data
	mov	temp, newout3
	rcall	shiftbyte
	mov	temp, newout2
	rcall	shiftbyte
	mov	temp, newout1
	rcall	shiftbyte
	;save the new display
	mov	oldout1, newout1
	mov	oldout2, newout2
	mov	oldout3, newout3
	rjmp	loop
	
date:
	;display date
.if (DD_MM_YY == 1)
	lds	newout1, RAM_Time+tDay
.else
	lds	newout1, RAM_Time+tYear
.endif
	lds	newout2, RAM_Time+tMonth
.if (DD_MM_YY == 1)
	lds	newout3, RAM_Time+tYear
.else
	lds	newout3, RAM_Time+tDay
.endif
	rjmp	display
	
time:
	;display time
	lds	newout1, RAM_Time+tHour
	lds	newout2, RAM_Time+tMinute
	mov	newout3, seconds
	rjmp	display
	
;*******************************************************************************
;*
;*   Nixie clock subroutines
;*
;*******************************************************************************

;---------------------------------------
; Shift out one byte (MSB first)
;   in: temp
;   modifies: temp, temp2
;---------------------------------------

shiftbyte:
	ldi	temp2, 8
shiftbyte_loop:
	cbi	PORTB, 6
	lsl	temp
	brcc	PC+2
	sbi	PORTB, 6
	sbi	PORTB, 7
	nop
	cbi	PORTB, 7
	dec	temp2
	brne	shiftbyte_loop
	ret

;---------------------------------------
; convert binary to packed BCD
;   in: temp
;   out: temp
;   modifies: temp2
;---------------------------------------

bin2bcd:
	ldi	temp2, 0
bin2bcd_loop:
	subi	temp2, -0x10
	subi	temp, 10
	brcc	bin2bcd_loop
	subi	temp2, 0x10
	subi	temp, -10
	add	temp, temp2
	ret
	